#!/usr/bin/perl -w
#
#  dBase Object
#  Use this module to access and write DBF files
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

use strict;
use FileHandle;
use Search;

package DBF;

#
#  Trim string
#
sub trim
{
   my ($str) = @_;
   if (defined($str))
   {
      $str =~ s/^  *//;
      $str =~ s/  *$//;
   }
   return $str;
}

#
#  Create a dBase object
#    $dbf = DBF->new("VAR1",...,"VARn");
#
sub new
{
   my $class = shift;
   my $self = {};
   bless($self,$class);

   my @var = @_;

   $self->erase();
   $self->{NVAR} = scalar(@var);
   my %var;
   for (my $k=0;$k<$self->{NVAR};$k++)
   {
      my ($name,$fmt) = split(/:/,$var[$k],2);
      exists($var{$name}) && die "Duplicate variable name $name";
      $var{$name}++;
      $self->{NAME}{$name} = $k;
      $self->{VAR}[$k]     = $name;
      $self->{FMT}[$k]     = $fmt;
   }

   return $self;
}

#
#  Create a dBase object by opening a file
#    $dbf = DBF->open("file");
#
sub open
{
   my $class = shift;
   my $self = {};
   bless($self,$class);

   my ($file,@opt) = @_;
   my $err = $self->Read($file,@opt);
   (!$err) || die "Error creating DBF $file: $err";

   return $self;
}

#
#  Read a DBF file
#    $error = $dbf->Read("file");
#
sub Read
{
   my $self = shift;
   my ($file,@opt) = @_;
   $self->erase();
   $self->{FILE} = $file;

   # Check the options
   my $direct = 0;
   my $uc     = 0;
   foreach my $opt (@opt)
   {
      if ($opt eq "DIRECT")
      {
         $direct = 1;
      }
      elsif ($opt eq "CACHE")
      {
         $direct = 0;
      }
      elsif ($opt eq "UC")
      {
         $uc = 1;
      }
      else
      {
         die "Unknown DBF option $opt";
      }
   }

   # Open dBase file and read header
   my $buffer;
   my $fd = new FileHandle($file);
   defined($fd) || return "Cannot open file $file";
   binmode($fd);
   (read($fd,$buffer,32)==32) || return "Cannot read header from $file";
   my ($magic,$yr,$mo,$dy,$nrec,$lhed,$lrec) = unpack("C4Vv2",$buffer);
   $yr += 1900;
   $self->{NREC} = $nrec;
   $self->{NVAR} = int($lhed/32)-1;

   # printf "Last Update:   %.2d/%.2d/%d\n" , $mo , $dy , $yr;
   # printf "Record Count:  %d\n" , $self->{NREC};
   # printf "Record Length: %d\n" , $lrec;
   # printf "Data offset:   %d\n" , $lhed;
   # printf "Variables:     %d\n" , $self->{NVAR};

   # Read variables
   my (@wid,@off,@m2a,$memo);
   for (my $k=0;$k<$self->{NVAR};$k++)
   {
      (read($fd,$buffer,32)==32) || return "Cannot read variable from $file";
      my ($name,$typ,undef,$wid,$dec) = unpack("A11a1VCC",$buffer);
      if ($typ eq "C")
      {
         $wid += 256*$dec;
         $dec = 0;
      }
      ($uc) && ($name = uc $name);
      if (exists($self->{NAME}{$name}))
      {
         for (my $l=1;;$l++)
         {
            unless (exists($self->{NAME}{"$name$l"}))
            {
               warn "Renamed $name to $name$l for uniqueness\n";
               $name .= $l;
               last;
            }
         }
      }
      $self->{VAR}[$k] = $name;
      $self->{NAME}{$name} = $k;
      $off[$k] = ($k==0) ? 1 : $off[$k-1]+$wid[$k-1];
      $wid[$k] = $wid;
      $m2a[$k] = ($typ eq "M") ? 1 : 0;
      ($typ ne "M") || ($memo = 1);
      if ($typ eq "C" || $typ eq "M")
      {
         $self->{FMT}[$k] = "A$wid";
      }
      elsif ($typ eq "L")
      {
         $self->{FMT}[$k] = "B$wid";
      }
      elsif ($typ eq "D")
      {
         $self->{FMT}[$k] = "D$wid";
      }
      elsif ($typ eq "N" || $typ eq "F")
      {
         $self->{FMT}[$k] = $wid . ($dec>0 ? ":$dec" : "");
      }
      else
      {
         return "Unknown variable type $file::$name : $typ : $wid : $dec";
      }
   }
   (read($fd,$buffer,1)==1 && (unpack("c",$buffer)==13 || unpack("c",$buffer)==10)) || return "Cannot read header terminus from $file : " . unpack("c",$buffer);

   my @memo;
   #  dBaseIV memo
   if ($magic == 0x8B)
   {
      !$direct || die "DIRECT invalid with memos";
      my $dbt = $file;
      $dbt =~ s/dbf$/dbt/;
      $dbt =~ s/DBF$/DBT/;
      my $fdt = new FileHandle($dbt);
      defined($fdt) || return "Cannot open file $dbt";
      binmode($fdt);
      for (my $k=0;read($fdt,$buffer,512)>0;$k++)
      {
         my ($tag,$len) = unpack("V2",$buffer);
         if ($tag == 0x0008FFFF)
         {
            $len -= 8;
            my $l = ($len<=504) ? $len : 504;
            $memo[$k] = substr($buffer,8,$l);
            for (my $dk=0 ; $len-=$l ; $dk++,$k++)
            {
               (read($fdt,$buffer,512)>0) || die "Error reading continuation block\n";
               $l = ($len<=512) ? $len : 512;
               $memo[$k-$dk] .= substr($buffer,0,$l);
            }
         }
      }
      $fdt->close();
      #  Clean up memos
      for (my $k=0;$k<@memo;$k++)
      {
         if (defined($memo[$k]))
         {
            $memo[$k] =~ s/\r/ /gs; #  CR and blanks
            $memo[$k] =~ s/\n/ /gs; #  NL and blanks
            $memo[$k] =~ s/^  *//;  #  Leading space
            $memo[$k] =~ s/  *$//;  #  Trailing space
            $memo[$k] =~ s/  */ /g; #  Extra space
            $memo[$k] =~ s/\0/ /g;  #  Nulls
         }
      }
   }
   elsif ($memo)
   {
      die "Not implemented:  memo for magic=$magic\n";
   }

   #  Set up for reads
   if ($direct)
   {
      $self->{FD} = $fd;
      $self->{LREC} = $lrec;
      @{$self->{OFF}} = @off;
      @{$self->{WID}} = @wid;
      #  Check for deleted records
      my $n=0;
      for (my $l=0;$l<$self->{NREC};$l++)
      {
         (read($fd,$buffer,1)==1) || return "Cannot read record from $file";
         if ($buffer ne '*')
         {
            $self->{DATA}[$n] = $n*$self->{LREC}+$lhed;
            $n++;
         }
         seek($fd,$lrec-1,IO::Seekable::SEEK_CUR) || return "Cannot seek $l";
      }
      my $term = (read($fd,$buffer,1)==1) ? unpack("c",$buffer) : 26;
      ($term==26) || return "Cannot read file terminus from $file : $term";
      $self->{NREC} = $n;
   }
   else
   {
      #  Read records
      my $n=0;
      for (my $l=0;$l<$self->{NREC};$l++)
      {
         (read($fd,$buffer,$lrec)==$lrec) || return "Cannot read record from $file";
         if (substr($buffer,0,1) ne '*')
         {
            for (my $k=0;$k<$self->{NVAR};$k++)
            {
               my $rec = substr($buffer,$off[$k],$wid[$k]);
               if ($m2a[$k])
               {
                  $rec =~ s/\0//g;
                  if ($rec eq "")
                  {
                  }
                  elsif (defined($memo[$rec]))
                  {
                     $rec = $memo[$rec];
                     (length($rec) <= $m2a[$k]) || ($m2a[$k] = length($rec));
                  }
                  else
                  {
                     die "Undefined memo $rec field $self->{VAR}[$k] record $n\n";
                     $rec = "";
                  }
               }
               else
               {
                  $rec =~ s/^  *//;
                  $rec =~ s/  *$//;
                  $rec =~ s/\0//g;
               }
               $self->{DATA}[$n][$k] = $rec;
            }
            $n++;
         }
      }
      my $term = (read($fd,$buffer,1)==1) ? unpack("c",$buffer) : 26;
      ($term==26) || return "Cannot read file terminus from $file : $term";
      $fd->close();
      $self->{NREC} = $n;
   
      #  Update memos
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         ($m2a[$k]==0) || ($self->{FMT}[$k] = "A$m2a[$k]");
      }
   }
   return;
}

#
#  Determine number of records
#    $size = $dbf->size();
#
sub size
{
   my $self = shift;
   return wantarray() ? ($self->{NREC},$self->{NVAR}) : $self->{NREC};
}

#
#  Rename DBF field
#    $dbf->rename("OLD","NEW:D8");
#
sub rename
{
   my $self = shift;
   (@_==2) || die "Wrong number of arguments for DBF->rename";
   my ($old,$new) = @_;
   my ($name,$fmt) = split(/:/,$new,2);
   exists($self->{NAME}{$old}) || die "Unknown old name $old in DBF->rename\n";
   my $k = $self->{NAME}{$old};
   delete $self->{NAME}{$old};
   $self->{NAME}{$name} = $k;
   $self->{VAR}[$k]     = $name;
   ($fmt) && ($self->{FMT}[$k] = $fmt);
}

#
#  Get field names or values
#    $dbf = DBF->fields('NAME');
#
sub fields
{
   my $self = shift;
   my ($what) = @_;

   if (!defined($what) || $what eq 'NAME')
   {
      return @{$self->{VAR}};
   }
   elsif ($what eq 'FORMAT')
   {
      my %fields;
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         $fields{$self->{VAR}[$k]} = $self->{FMT}[$k];
      }
      return %fields;
   }
   elsif ($what eq 'WIDTH')
   {
      my %fields;
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         my ($wid,$dec) = split(/:/,$self->{FMT}[$k]);
         $wid =~ s/[A-Z]//;
         $fields{$self->{VAR}[$k]} = $wid;
      }
      return %fields;
   }
   elsif ($what eq 'PRINTF')
   {
      my %fields;
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         my ($wid,$dec) = split(/:/,$self->{FMT}[$k]);
         my $typ = ($wid =~ /[A-Z]/) ? substr($wid,0,1) : "";
         $wid =~ s/[A-Z]//;
         $fields{$self->{VAR}[$k]} = ($typ eq "A") ? "%-${wid}s" : "%${wid}s";
      }
      return %fields;
   }
   elsif ($what eq 'DBF')
   {
      my @fields;
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         $fields[$k] = $self->{VAR}[$k].':'.$self->{FMT}[$k];
      }
      return @fields;
   }
   else
   {
      die "Unknown fields type $what";
   }
}

#
#  Make sure database is strictly correct
#
sub strict
{
   my $self = shift;
   for (my $k=0;$k<$self->{NVAR};$k++)
   {
      my $mc = $self->{VAR}[$k];
      my $uc = uc($mc);
      if ($uc ne $mc)
      {
         !exists($self->{NAME}{$uc}) || die "Conflicting variable names $mc and $uc\n";
         $self->{NAME}{$uc} = $k;
         delete $self->{NAME}{$mc};
         $self->{VAR}[$k] = $uc;
      }
   }
}

#
#  Determine variable format from data
#    $dbf->format($force);
#
sub format
{
   my $self = shift;
   my ($force,%opt) = @_;

   if (uc $force eq 'FREE')
   {
      $self->{FMT} = ();
      return;
   }

   my @kill = exists($opt{KILL}) ? @{$opt{KILL}} : ();

   for (my $k=0;$k<$self->{NVAR};$k++)
   {
      if ($force || !defined($self->{FMT}[$k]))
      {
         my $d8  = (defined($self->{FMT}[$k]) && substr($self->{FMT}[$k],0,1) eq "D");
         my $nil = 0;
         my $nan = 0;
         my $ast = 0;
         my $wid = 0;
         my $int = 0;
         my $dec = 0;
         my $typ = "N";
         for (my $n=0;$n<$self->{NREC};$n++)
         {
            foreach my $kill (@kill)
            {
               $self->subst($n,$k,$kill,'');
            }
            my $var = $self->data($n,$k);
            (defined $var) || die "Undefined field $self->{VAR}[$k] : $n";
            ($wid < length($var)) && ($wid = length($var));
            # For numbers figure decimals
            if ($typ eq "N")
            {
               #  I can live with NaN and "" and ****
               if ($var eq "")
               {
                  $nil++;
                  $nan++;
               }
               elsif (uc($var) eq "NAN")
               {
                  $nan++;
               }
               elsif ($var =~ /^\*+$/)
               {
                  $ast++;
                  $nan++;
               }
               #  See if it is an integer
               elsif ($var =~ s/^\s*([+-]?\d+)\s*$/$1/)
               {
                   ($int < length($var)) && ($int = length($var));
               }
               #  See if it is a decimal number
               elsif ($var =~ s/^\s*([+-]?\d*\.\d*)\s*$/$1/)
               {
                  my ($i,$d) = split(/\./,$var);
                  (defined($i) && $int < length($i)) && ($int = length($i));
                  if (defined($d))
                  {
                     $d =~ s/0*$//;  #  Kill trailing zeros
                     ($dec < length($d)) && ($dec = length($d));
                  }
               }
               #  See if it is in exponential format
               elsif ($var =~ /^\s*([+-]?\d*\.\d*)[de]([+-]?\d+)\s*$/)
               {
                  my ($mant,$exp) = ($1,$2);
                  if (abs($exp)>20)
                  {
                     $typ = "C";
                  }
                  else
                  {
                     my ($i,$d) = split(/\./,$mant);
                     (defined($i) && $int < length($i)+$exp) && ($int = length($i)+$exp);
                     if (defined($d))
                     {
                        $d =~ s/0*$//;  #  Kill trailing zeros
                        ($dec < length($d)-$exp) && ($dec = length($d)-$exp);
                     }
                  }
               }
               #  Gotta be a string (or bool)
               else
               {
                  $typ = "C";
               }
            }
         }
         $int = 3 if ($typ eq "N" && $nan>0 && $int<3);
         #  Everything is blank or everything ***
         if ($nil==$self->{NREC} || $ast==$self->{NREC})
         {
            $typ = "C";
            $self->{FMT}[$k] = 'A'.($wid>1?$wid:1);
         }
         #  non-numeric
         elsif ($typ ne "N")
         {
            $self->{FMT}[$k] = "A$wid";
         }
         #  Very long integers stored as doubles
         elsif ($int>9 && $dec==0)
         {
            $self->{FMT}[$k] = sprintf "%d:0" , $int;
         }
         #  float/double
         elsif ($dec>0 || $int>9)
         {
            $self->{FMT}[$k] = sprintf "%d:%d" , $int+1+$dec , $dec;
         }
         #  date
         elsif ($d8 && $int==8)
         {
            $self->{FMT}[$k] = "D$int";
         }
         #  integer
         else
         {
            ($int<1) && ($int=1);  #  Strange things like all .000
            $self->{FMT}[$k] = "$int";
         }
      }
   }
   return;
}

#
#  Write a DBF file
#
sub write
{
   my $self = shift;
   my ($file,@opt) = @_;
   $file =~ s/^>*//;
   $self->format(0);

   #  Check options
   my $NoNaN = 0;
   foreach my $opt (@opt)
   {
      if ($opt eq "NoNaN")
      {
         $NoNaN = 1;
      }
      else
      {
         die "Unknown DBF->write option $opt\n";
      }
   }

   #  Set up variables
   my (@typ,@wid,@dec);

   #  Set up dBase format
   my (@fmt,@len);
   for (my $k=0;$k<$self->{NVAR};$k++)
   {
      my ($wid,$dec) = split(/:/,$self->{FMT}[$k]);
      $len[$k] = $wid[$k] = $wid;
      $dec[$k] = (defined($dec)) ? $dec : 0;
      if ($wid =~ /^\d+$/)
      {
         $typ[$k] = "N";
      }
      elsif (substr($wid,0,1) eq "D")
      {
         $typ[$k] = "D";
         $len[$k] = $wid[$k] = substr($wid,1);
      }
      elsif (substr($wid,0,1) eq "A")
      {
         $typ[$k] = "C";
         $len[$k] = substr($wid,1);
         $wid[$k] = $len[$k] % 256;
         $dec[$k] = int($len[$k] / 256);
      }
      elsif (substr($wid,0,1) eq "B")
      {
         $typ[$k] = "L";
         $len[$k] = $wid[$k] = substr($wid,1);
      }
      if ($typ[$k] eq "C")
      {
         $fmt[$k] = sprintf "%%-%ds" , $len[$k];
      }
      elsif ($typ[$k] eq "B")
      {
         $fmt[$k] = sprintf "%%%ds" , $wid[$k];
      }
      elsif ($typ[$k] ne "N")
      {
         $fmt[$k] = sprintf "%%-%ds" , $wid[$k];
      }
      elsif ($dec[$k] == 0)
      {
         $fmt[$k] = "%$wid[$k]d";
      }
      else
      {
         $fmt[$k] = "%$wid[$k].$dec[$k]f";
      }
   }

   #  Compute header variables
   my $magic = 3;
   my (undef,undef,undef,$dy,$mo,$yr) = localtime(); $mo++;
   my $lhed = $self->{NVAR}*32+33;
   my $lrec = 1;
   for (my $k=0;$k<$self->{NVAR};$k++)
   {
      $lrec += $len[$k];
   }

   #  Write dBASE file
   my $fd = new FileHandle(">$file");
   defined($fd) || die "Cannot open file $file";
   binmode($fd);
   print $fd pack("C4Vv2a20" , $magic , $yr , $mo , $dy , $self->{NREC} , $lhed , $lrec);
   for (my $k=0;$k<$self->{NVAR};$k++)
   {
      print $fd pack("a11a1a4CCa14" , $self->{VAR}[$k] , $typ[$k] , "" , $wid[$k] , $dec[$k]);
   }
   print $fd pack("C" , 13);
   for (my $n=0;$n<$self->{NREC};$n++)
   {
      print $fd " ";
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         my $var = $self->data($n,$k);
         if ($typ[$k] =~ /[ND]/ && (uc($var) eq "NAN" || $var eq "" || $var =~ /^\**$/))
         {
            $var = sprintf "%$len[$k]s" , ($NoNaN ? "" : "NaN");
         }
         elsif ($typ[$k] eq "N" && $var eq ".")
         {
            $var = sprintf $fmt[$k] , 0;
         }
         else
         {
            $var = sprintf $fmt[$k] , $var;
         }
         (length($var)==$len[$k]) || die "Field overflow $len[$k]: $self->{VAR}[$k]=$var";
         print $fd $var;
      }
   }
   print $fd pack("C" , 26);
   $fd->close();
}

#
#  Erase a dBase object
#
sub erase
{
   my $self = shift;

   #  This helps the garbage collector
   my $N = exists($self->{NREC}) ? $self->{NREC}: 0;
   for (my $n=0;$n<$N;$n++)
   {
      $self->{DATA}[$n] = ();
   }
   #  Initialize variables
   $self->{NREC} = 0;
   $self->{NVAR} = 0;
   $self->{NAME} = ();
   $self->{VAR}  = ();
   $self->{FMT}  = ();
   $self->{DATA} = ();
   $self->undex();
   #  Delete variables used for DIRECT
   ($self->{FD}) && $self->{FD}->close();
   delete $self->{FD};
   delete $self->{LREC};
   delete $self->{OFF};
   delete $self->{WID};
}

#
#  Delete all records in dBase object
#
sub clear
{
   my $self = shift;

   #  This helps the garbage collector
   for (my $n=0;$n<$self->{NREC};$n++)
   {
      $self->{DATA}[$n] = ();
   }
   #  Update records
   $self->{NREC} = 0;
   $self->{DATA} = ();
   $self->undex();
}

#
#  Add a record to database
#    Call as $dbf->add({VAR1 => VAL1 , ... , VARn => VALn})
#         or $dbf->add(VAL1 , ... , VALn)
#         or $dbf->add{$dbf0);
#
sub add
{
   my $self = shift;
   my $n = $self->{NREC};
   ($self->{FD}) && die "Cannot add a DIRECT DBF";
   #  Database Union
   if (@_==1 && ref($_[0]) eq "DBF")
   {
      my ($db) = @_;
      ($self->{NVAR} == $db->{NVAR}) || die "Number of variables differ in DBF add";
      my @k;
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         (exists($db->{NAME}{$self->{VAR}[$k]})) || die "Variable $self->{VAR}[$k] missing in DBF add";
         $k[$k] = $db->{NAME}{$self->{VAR}[$k]};
      }
      for (my $i=0;$i<$db->{NREC};$i++)
      {
         for (my $k=0;$k<$self->{NVAR};$k++)
         {
            $self->{DATA}[$n][$k] = $db->{DATA}[$i][$k[$k]];
         }
         $n++;
      }
   }
   #  Hashed record
   elsif (@_==1 && ref($_[0]) eq "HASH")
   {
      my ($rec) = @_;
      foreach my $name (keys %{$rec})
      {
         exists($self->{NAME}{$name}) || die "Unknown key for hash add $name";
         my $k = $self->{NAME}{$name};
         $self->{DATA}[$n][$k] = defined($rec->{$name}) ? trim($rec->{$name}) : "";
      }
      $n++;
   }
   #  Array record
   else
   {
      (@_==$self->{NVAR}) || die "Wrong number of arguments for array add";
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         $self->{DATA}[$n][$k] = defined($_[$k]) ? trim($_[$k]) : "";
      }
      $n++;
   }
   #  Update count
   $self->{NREC} = $n;
}

#
#  Delete a record to database
#    Call as $dbf->delete($n);
sub delete
{
   my $self = shift;
   (@_==1) || die "Wrong argument count in DBF->delete\n";
   my ($N) = @_;
   (0<=$N && $N<$self->{NREC}) || die "DBF->delete record out of range $N\n";
   #  Update records
   $self->{NREC}--;
   for (my $n=$N;$n<$self->{NREC};$n++)
   {
      $self->{DATA}[$n] = $self->{DATA}[$n+1];
   }
}

#
#  Perform substitute on a field in the database
#
sub subst
{
   my $self = shift;
   my ($n,$k,$src,$dst,$opt) = @_;

   ($self->{FD}) && die "Cannot do subst on file\n";
   (0<=$n && $n<$self->{NREC}) || die "Invalid record in subst $n\n";
   (0<=$k && $k<$self->{NVAR}) || die "Invalid field in subst $k\n";
   defined($self->{DATA}[$n][$k]) || return;
   $self->{DATA}[$n][$k] =~ s/$src/$dst/g;
}

#
#  Get a field from database
#  Returns empty if record out of range
#
sub data
{
   my $self = shift;
   my ($n,$k) = @_;

   if (!defined($k))
   {
      my $rec;
      seek($self->{FD},$self->{DATA}[$n],IO::Seekable::SEEK_SET) || return "Cannot seek $n";
      (read($self->{FD},$rec,$self->{LREC})==$self->{LREC}) || return "Cannot read $n";
      return $rec;
   }
   elsif ($self->{FD})
   {
      my $rec;
      seek($self->{FD},$self->{DATA}[$n]+$self->{OFF}[$k],IO::Seekable::SEEK_SET) || return "Cannot seek $n:$k";
      (read($self->{FD},$rec,$self->{WID}[$k])==$self->{WID}[$k]) || return "Cannot read $n:$k";
      $rec =~ s/^  *//;
      $rec =~ s/  *$//;
      $rec =~ s/\0//g;
      return $rec;
   }
   else
   {
      return $self->{DATA}[$n][$k];
   }
}

#
#  Get a record from database
#  Returns empty if record out of range
#
sub get
{
   my $self = shift;
   my ($n,@var) = @_;
   if (0<=$n && $n<$self->{NREC})
   {
      if (@var==0)
      {
         my %rec;
         for (my $k=0;$k<$self->{NVAR};$k++)
         {
            $rec{$self->{VAR}[$k]} = $self->data($n,$k);
         }
         return %rec;
      }
      elsif (wantarray())
      {
         my @rec;
         foreach my $var (@var)
         {
            my $k = $self->{NAME}{$var};
            defined($k) || die "Invalid variable $var\n";
            push @rec , $self->data($n,$k);
         }
         return @rec;
      }
      else
      {
         my $var = $var[0];
         my $k = $self->{NAME}{$var};
         defined($k) || die "Invalid variable $var\n";
         return $self->data($n,$k);
      }
   }
}

#
#  Find record(s) from database
#  Returns list of matching records
#  Call with @reclist = $dbf->find("FIELD_NAME","SEARCH_VALUE","TYPE OF MATCH");
#  TYPE is A for strings, U for stings case independent or N for numeric
#  TYPE ~A, ~U and ~N returns all the records that does NOT match
#
sub find
{
   my $self = shift;
   (@_==3) || die "Wrong number of arguments in DBF->find";
   my ($var,$val,$type) = @_;
   exists($self->{NAME}{$var}) || die "Unknown variable $var in find";
   my $k = $self->{NAME}{$var};
   my @n;
   if ($k == $self->{IDXVK} && $self->{IDXTP} eq $type)
   {
      if ($type ne "N")
      {
         exists($self->{INDEX}{$val}) && (@n = @{$self->{INDEX}{$val}});
      }
      elsif (uc($val) eq "NAN")
      {
         @n = @{$self->{INDEX}{NaN}};
      }
      else
      {
         my $k = Search::binfind($val,$self->{INDEX}{X},"N");
         defined($k) && (@n = @{$self->{INDEX}{N}[$k]});
      }
   }
   elsif ($type eq "A")
   {
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         ($self->data($n,$k) ne $val) || (push @n , $n);
      }
   }
   elsif ($type eq "U")
   {
      $val = uc($val);
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         (uc($self->data($n,$k)) ne $val) || (push @n , $n);
      }
   }
   elsif ($type eq "N")
   {
      if (uc($val) eq "NaN")
      {
         for (my $n=0;$n<$self->{NREC};$n++)
         {
            ($self->data($n,$k) ne "NaN") || (push @n , $n);
         }
      }
      else
      {
         for (my $n=0;$n<$self->{NREC};$n++)
         {
            my $data = $self->data($n,$k);
            ($data eq "NaN") || ($data != $val) || (push @n , $n);
         }
      }
   }
   elsif ($type eq "~A")
   {
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         ($self->data($n,$k) eq $val) || (push @n , $n);
      }
   }
   elsif ($type eq "~U")
   {
      $val = uc($val);
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         (uc($self->data($n,$k)) eq $val) || (push @n , $n);
      }
   }
   elsif ($type eq "~N")
   {
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         ($self->data($n,$k) == $val) || (push @n , $n);
      }
   }
   else
   {
      die "Unknown find type $type";
   }
   return @n
}

#
#  Look up a field in the database
#  Returns a single value or array depending on context
#  Call with @vals = $dbf->lookup("SOURCE_FIELD","SEARCH_VALUE","TYPE OF MATCH","TAR1",...,TARn);
#
sub lookup
{
   my $self = shift;
   (@_>=3) || die "Wrong number of arguments in DBF->lookup";
   my ($var,$val,$type,@tar) = @_;
   #  Get index of requested variables
   my @k;
   foreach my $tar (@tar)
   {
      exists($self->{NAME}{$tar}) || die "Unknown target $tar in lookup of $self->{FILE}";
      push @k , $self->{NAME}{$tar};
   }
   my @n = $self->find($var,$val,$type);
   #  When lookup is not unique, check if the results are
   if (@n>1)
   {
      my $n0 = shift @n;
      if (@tar>0)
      {
         foreach my $n (@n)
         {
            foreach my $k (@k)
            {
               ($self->data($n0,$k) eq $self->data($n,$k)) || die "DBF->lookup is not unique for $var=$val: $n0 @n\n";
            }
         }
      }
      else
      {
         foreach my $n (@n)
         {
            for (my $k=0;$k<$self->{NVAR};$k++)
            {
               ($self->data($n0,$k) eq $self->data($n,$k)) || die "DBF->lookup is not unique for $var=$val: $n0 @n\n";
            }
         }
      }
      #  Result is degenerate
      @n = ($n0);
   }
   #  Nothing found
   if (@n==0)
   {
      return;
   }
   #  One found
   elsif (@tar==0)
   {
      my %rec;
      for (my $k=0;$k<$self->{NVAR};$k++)
      {
         $rec{$self->{VAR}[$k]} = $self->data($n[0],$k);
      }
      return %rec;
   }
   elsif (wantarray())
   {
      my @x;
      foreach my $k (@k)
      {
         push @x , $self->data($n[0],$k);
      }
      return @x;
   }
   else
   {
      return $self->data($n[0],$k[0]);
   }
}

#
#  Return a subset of records and fields from the database
#  Returns an array for hashes
#  Call with @vals = $dbf->select("SOURCE_FIELD","SEARCH_VALUE","TYPE OF MATCH","TAR1",...,"TARn");
#
sub select
{
   my $self = shift;
   (@_>=3) || die "Wrong number of arguments in DBF->select";
   my ($var,$val,$type,@tar) = @_;
   foreach my $tar (@tar)
   {
      exists($self->{NAME}{$tar}) || die "Unknown variable $tar in select";
   }
   (@tar==0) && (@tar = @{$self->{VAR}});
   #  Find records
   my @n = $self->find($var,$val,$type);
   #  Set up return
   my @sel;
   for (my $i=0;$i<@n;$i++)
   {
      foreach my $tar (@tar)
      {
         $sel[$i]{$tar} = $self->data($n[$i],$self->{NAME}{$tar});
      }
   }
   return @sel;
}

#
#  Destroy index
#  Does not return a value
#
sub undex
{
   my $self = shift;
   $self->{IDXTP} = "";
   $self->{IDXVK} = -1;
   $self->{INDEX} = ();
}

#
#  Index database on a field
#  This speeds up access on that field with "find"
#  Does not return a value
#
sub index
{
   my $self = shift;
   (@_==2) || die "Wrong number of arguments in DBF->index";
   my ($var,$type) = @_;
   exists($self->{NAME}{$var}) || die "Unknown variable $var in index";
   my $k = $self->{NAME}{$var};
   $self->{IDXVK} = $k;
   $self->{IDXTP} = $type;
   $self->{INDEX} = ();
   if ($type eq "A")
   {
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         push @{$self->{INDEX}{$self->data($n,$k)}} , $n;
      }
   }
   elsif ($type eq "U")
   {
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         push @{$self->{INDEX}{uc($self->data($n,$k))}} , $n;
      }
   }
   elsif ($type eq "N")
   {
      #  Assemble index field - trap nan's
      my $i=0;
      my @idx;
      $self->{INDEX}{NaN} = ();
      $self->{INDEX}{X}   = ();
      $self->{INDEX}{N}   = ();
      for (my $n=0;$n<$self->{NREC};$n++)
      {
         my $val = $self->data($n,$k);
         if ($val eq "NaN")
         {
            push @{$self->{INDEX}{NaN}} , $n;
         }
         else
         {
            $idx[$i]{X} = $val;
            $idx[$i]{N} = $n;
            $i++;
         }
      }
      #  Store on index field
      if (@idx>0)
      {
         @idx = sort {$a->{X} <=> $b->{X}} @idx;
         $i=0;
         $self->{INDEX}{X}[$i]    = $idx[0]{X};
         $self->{INDEX}{N}[$i][0] = $idx[0]{N};
         for (my $n=1;$n<@idx;$n++)
         {
            if ($self->{INDEX}{X}[$i] == $idx[$n]{X})
            {
               push @{$self->{INDEX}{N}[$i]} , $idx[$n]{N};
            }
            else
            {
               $i++;
               $self->{INDEX}{X}[$i]    = $idx[$n]{X};
               $self->{INDEX}{N}[$i][0] = $idx[$n]{N};
            }
         }
      }
   }
   else
   {
      die "Invalid index type $type";
   }
}

#
#  Count occurences in database
#  Returns hash of value => occurences
#  Call with %reclist = $dbf->count("FIELD_NAME");
#
sub count
{
   my $self = shift;
   (@_==1) || die "Wrong number of arguments in DBF->count";
   my ($var) = @_;
   exists($self->{NAME}{$var}) || die "Unknown variable $var in count";
   my $k = $self->{NAME}{$var};
   my %n;
   for (my $n=0;$n<$self->{NREC};$n++)
   {
      $n{$self->data($n,$k)}++;
   }
   return %n
}

#
#  DBF destructor
#
sub DESTROY
{
   my $self = shift;
   #  Calling erase helps the garbage collector
   $self->erase();
}

#
#  Join databases
#    Call as $dbf->join(FIELD,$jdb,FIELD,TYPE,VAR1, ... , VARn)
#
sub join
{
   my $self = shift;
   ($self->{FD}) && die "Cannot join a DIRECT DBF";
   #  Check data
   my ($tar,$dbf,$src,$type,@var) = @_;
   exists($self->{NAME}{$tar}) || die "Variable $tar not in database\n";
   my $k = $self->{NAME}{$tar};
   exists($dbf->{NAME}{$src}) || die "Variable $src not in database\n";
   #  Add variables
   (@var>0) || die "No variables to join\n";
   my $k0 = $self->{NVAR};
   for (my $i=0;$i<@var;$i++)
   {
      exists($dbf->{NAME}{$var[$i]})   || die "Variable $var[$i] not in database\n";
      my $j = $dbf->{NAME}{$var[$i]};
      !exists($self->{NAME}{$var[$i]}) || die "Variable $var[$i] already in database\n";
      my $k = $self->{NVAR}++;
      $self->{VAR}[$k] = $dbf->{VAR}[$j];
      $self->{FMT}[$k] = $dbf->{FMT}[$j];
      $self->{NAME}{$var[$i]} = $k;
   }
   for (my $n=0;$n<$self->{NREC};$n++)
   {
      my $val  = $self->data($n,$k);
      my @vals = $dbf->select($src,$val,$type,@var);
      #  Missing data is set to a blank
      if (@vals==0)
      {
         for (my $i=0;$i<@var;$i++)
         {
            $self->{DATA}[$n][$k0+$i] = "";
         }
      }
      #  Unique match
      elsif (@vals==1)
      {
         for (my $i=0;$i<@var;$i++)
         {
            $self->{DATA}[$n][$k0+$i] = $vals[0]{$var[$i]};
         }
      }
      else
      {
         die "Value $val is not unique in join\n";
      }
   }
}

#
#  Read DBF file into a hash
#
sub ReadHash
{
   my ($file,$src,$tar) = @_;
   my $dbf = DBF->open($file,"DIRECT");
   my %hash;
   my $n = $dbf->size();
   for (my $i=0;$i<$n;$i++)
   {
      my ($key,$val) = $dbf->get($i,$src,$tar);
      $hash{$key} = $val;
   }
   return %hash;
}

1;
